from .numpy_vector import NumpyPrimalVector
from .numpy_vector import NumpyDualVector

from .dolfin_vector import DolfinPrimalVector
from .dolfin_vector import DolfinDualVector
from .dolfin_vector import RieszMap



from .dolfin_vector_set import DolfinPrimalVectorSet
from .dolfin_vector_set import DolfinDualVectorSet
from .dolfin_vector_set import RieszMapSet

from .adaptor import convert_to_moola_dual_vector